//
//    LineFind produces searches fo line featyres in the spectrum of 
//    a channel. 
//    The configuration for LineFind is specified by an ascii file 
//    passed to LineFind() as an argument. The file contains a single
//    line:
//
//    Line 1: <channel> <tAverage> 
//
//    The fields have the following meaning:
//    <channel>      The name of the channel from which the data are derived.
//    <tAverage>     The averaging time over which the spectrum is taken.
//    
//
//

#include "DLineFind.hh"

#ifndef __CINT__
Dacc In("/relativity/LIGO1/H-609940/*.F");
#endif

//--------------------------------------  Channel descriptor class
class Chan {
public:
  Chan() {}
  ~Chan() {}
  char mName[40];
  double mRate;
};

class DLineFind : public DatEnv {
  public:
  DLineFind(int argc, const char* argv[]);
  ~DLineFind();
  void ProcessData(void);
  private:
  Chan chanv;
  int nWin;      // number of taper windows to use
  int nPoint;
  int paddingFactor;
  int paddedLength;
  int maxLines;  // maximum number of lines to remove
  int nFreq;
  int nRemoved;
  float nwdt;    // time-freq bandwidth 
  float *data;
  float *initial_data;
  float *mtap_spec_init;
  float *mtap_spec_final;
  double tsecs;  // sample time in seconds
  RemovedLines *line_list;
  TCanvas *c;
  bool writeFiles;
  bool snapshot;
  bool first;
  ofstream fpout1;
  ofstream fpout2;
  const float fpi;
};

using namespace std;

void
LineFind(const char* config) {
    DLineFind xxx(1, &config);
    xxx.MainLoop();
}

//--------------------------------------  Define the main program in background
#ifndef __CINT__
// equivalent to EXEROOT - see DatEnv.hh
TROOT root("DatEnv", "Graphical background monitor"); 
int main(int argc, const char* argv[]) { 
    TApplication theApp("ACO Multitaper Line Monitor", &argc, const_cast<char**>(argv)); 
    DLineFind MyObject(argc,argv); 
    MyObject.MainLoop();
    return 0;}   
#endif


DLineFind::DLineFind(int argc, const char* argv[]) 
  :DatEnv(argc, argv), nWin(5), paddingFactor(4), maxLines(100), nwdt(3.0), 
  writeFiles(false), snapshot(false), fpi(M_PI)
 {
   const char* config;
   config = argv[argc-1];
   if (!config || !*config) {
     cerr<< "# LineFind: configuration file not specified" << endl;
     return;
   }

   //---------------------------------  Get parameters.
   ifstream ifs(config, ios::in);
   if (!ifs) {
     cerr << "# LineFind: unable to open configuration file " << config << endl;
     return;
   }

   cerr << "# LineFind: reading parameters from config file: " << config << endl;
   
   ifs >> chanv.mName >> tsecs; 
   cerr << "# LineFind: using channel " << chanv.mName << " with sample time = " 
	   << tsecs << " seconds" <<endl;
   ifs.close();

   mIn.addChannel(chanv.mName);  //  Set channel name
   mIn.setBuffer(0);       // Set to take up data as requested
   mIn.setStride(tsecs);   // No of seconds of data to take up

   // initialise
   first=true;
   // initialise ROOT graphics
   gROOT->Reset();
   c = new TCanvas("c", "Spectrum before and after line removal",10,10,788,750);
   gStyle->SetOptStat(0);   //   Don't show statistics
   gStyle->SetOptTitle(0);  //   Don't show individual tiles -- we'll add a global title
   gROOT->SetInterrupt(kTRUE); 

}
 
void
DLineFind::ProcessData() {    
  //-----------  Calculate length of time interval of data required
  float freq,amp,phi,*x;
  
  if (first) {  // We must determine the sample rate and 
                // allocate storage on the first pass
    {
	 char UTCdate[256];
	 TimeStr((mIn.refData(chanv.mName))->getStartTime(), UTCdate,"%H:%N:%S %M %d %Y %Z");
	 cerr << "# LineFind: starting at time " << UTCdate << endl;
    }

    chanv.mRate=1./((mIn.refData(chanv.mName))->getTStep());  // sample rate
	
    nPoint=int(tsecs*chanv.mRate);  // total number of data points
    paddedLength=paddingFactor*nPoint;   // number of points with zero pading
    nFreq=1+paddedLength/2;     // num frequencies including DC, Nyquist 
   
    cerr << "nPoint: " << nPoint << "   paddedLength: " << "     nFreq: " << endl;
    // allocate arrays
    cerr << "allocating arrays" << endl;
    data=new float[nPoint];
    clear(data,nPoint,1);    // zero the array
    initial_data=new float[nPoint];
    mtap_spec_init=new float[nFreq];
    mtap_spec_final=new float[nFreq];
    cerr << "allocated arrays" << endl;
    cerr << "allocating line_list" << endl;
    line_list=new RemovedLines[maxLines];
     cerr << "allocated line_list" << endl;
   
    first=false;
  }

   // check we have the number of points we expect
   int ptsRetd=(mIn.refData(chanv.mName))->getData(nPoint,data);
   if (ptsRetd != nPoint) {
	cerr << "LineFind: Points requested = " << nPoint << 
	  ",  Points returned = " << ptsRetd <<endl;
	abort();
   }

  // subtract the mean value from the data set 
  remove_mean(data,nPoint);
  // save initial data set  
  for (int i=0;i<nPoint;i++) initial_data[i]=data[i];

  // remove the spectral lines from the data set 
  remove_spectral_lines(data,nPoint,paddedLength,nwdt,nWin,
     maxLines,500,&nRemoved,line_list,mtap_spec_init,mtap_spec_final,1,0,nFreq);
    
  // print out a list of lines foundd 
  printf("Total number of lines found: %d\n",nRemoved);
  for (int i=0;i<nRemoved;i++) {
    freq=0.5*line_list[i].index*chanv.mRate/nFreq;
    amp=2.0*sqrt(line_list[i].re*line_list[i].re+line_list[i].im*line_list[i].im);
    phi=180*atan2(line_list[i].im,line_list[i].re)/fpi;
    cerr << "Found line frequency " << freq << "Hz amplitude " << amp << " phase "
	 << phi << " (F-test " << line_list[i].fvalue  << ")" << endl;
  }

  x=new float[nFreq];
  for (int i=0;i<nFreq;i++) {
    x[i]=0.5*i*chanv.mRate/nFreq;
  }    

  // now output a file containing the initial and final data... 
  if (writeFiles) {
    fpout1.open("ifo_clean_data.out", ios::out);
    if ( fpout1.bad() ) {
	 cerr << "LineFind: Cannot open ifo_clean_data.out" << endl;
	 finish();
	 return;
    } 
    cerr << "LineFind: Writing ifo_clean_data.out" << endl;

    fpout1 << "# Three columns are: " << endl;
    fpout1 << "# Time (sec) \t Initial data \t Final Data" << endl;
    for (int i=0;i<nPoint;i++)
	 fpout1 << i/chanv.mRate << "\t" <<  initial_data[i] << "\t" <<  data[i] << endl;

    // ... and the initial and final spectra

    fpout2.open("ifo_clean_spec.out", ios::out);
    if ( fpout2.bad() ) {
	 cerr << "LineFind: Cannot open ifo_clean_spec.out" << endl;
	 finish();
	 return;
    } 
    cerr << "LineFind: Writing ifo_clean_spec.out" << endl;
    
    fpout2 << "# Three columns are:" << endl;
    fpout2 << "# Freq (Hz) \t  Initial spectrum  \t Final spectrum" << endl;
    for (int i=0;i<nFreq;i++) {
	 fpout2 << x[i] << "\t" << mtap_spec_init[i] << "\t" 
		   << mtap_spec_final[i] << endl;
    }

  }

	
  c->SetHighLightColor(2);
  c->SetFillColor(10); // White in the default palette - see TROOT::GetListOfColors 
  c->SetBorderSize(2);
  // c->SetLogx();
  c->SetLogy();
  c->SetGridx();
  c->SetGridy();
  c->SetTickx();
  c->SetTicky();
  c->SetLeftMargin(0.12);
  c->SetRightMargin(0.0);
  c->SetTopMargin(0.08);
  c->SetBottomMargin(0.12);
  // We include a dummy histogram to overcome a bug in ROOT:
  // when the first histogram is continuous , Draw("C"), there is no output
  TH1F *dummy  = new TH1F("dummy","Dummy",nFreq,x[0],x[nFreq-1]);
  TH1F *before  = new TH1F("before","Initial Spectrum",nFreq,x[0],x[nFreq-1]);
  TH1F *after = new TH1F("after","Modified Spectrum",nFreq,x[0],x[nFreq-1]);

  for (int i=0;i<nFreq;i++) {
	  dummy->Fill(x[i],mtap_spec_init[i]);
	  before->Fill(x[i],mtap_spec_init[i]);
	  after->Fill(x[i],mtap_spec_final[i]);
	}                            

  dummy->SetLineColor(10); // White
  dummy->Draw();           // Draw a histogram  

  before->SetLineWidth(1);
  before->SetLineColor(4);     // Blue in the default ROOT palette
  before->Draw("CSAME");       // Draw a continuous line between points
  before->SetXTitle("Frequency (Hz)");

  after->SetLineColor(2);    // Red in the default ROOT palette 
  after->SetLineWidth(1);
  after->Draw("CSAME");
	
  c->RedrawAxis();           //  Redraw so axes appears above lines 	

  TPaveText *pt = new TPaveText(0.3, 0.935, 0.7, 0.99,"NDC");
  pt->SetFillColor(18);
  pt->AddText(chanv.mName);
  pt->Draw();

  c->Update();

  if (snapshot) {
    finish();  //  just process a single time segment
  }
}


DLineFind::~DLineFind() {
  //---------------------------------  Clean up at the end
  In.rmChannel(chanv.mName);
  if (writeFiles) {
    fpout1.close();
    fpout2.close();
  }
 
  cout << "LineFind finished" << endl;
}               
